# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import json
import sys

import numpy
import pytest
import skhep_testdata

import uproot

awkward = pytest.importorskip("awkward")


def test_awkward_strings():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["string"].array(library="ak")) == [
            "one",
            "two",
            "three",
            "four",
            "five",
        ]


def test_awkward_tstring():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["tstring"].array(library="ak")) == [
            "one",
            "two",
            "three",
            "four",
            "five",
        ]


def test_awkward_vector_int32():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["vector_int32"].array(library="ak")) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]


def test_awkward_vector_string():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["vector_string"].array(library="ak")) == [
            ["one"],
            ["one", "two"],
            ["one", "two", "three"],
            ["one", "two", "three", "four"],
            ["one", "two", "three", "four", "five"],
        ]


def test_awkward_vector_string_forth():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        temp_branch = tree["vector_string"]
        temp_branch.interpretation._forth = True
        assert awkward.to_list(temp_branch.array(library="ak")) == [
            ["one"],
            ["one", "two"],
            ["one", "two", "three"],
            ["one", "two", "three", "four"],
            ["one", "two", "three", "four", "five"],
        ]


def test_awkward_array_tref_array_forth():

    awk_data = None

    with uproot.open(skhep_testdata.data_path("uproot-delphes-pr442.root"))[
        "Delphes"
    ] as tree:
        temp_branch = tree["GenJet04/GenJet04.Constituents"]
        temp_branch.interpretation._forth = True
        awk_data = temp_branch.array(library="ak")

    assert awk_data[0][0]["refs"][-1] == 2579
    assert awk_data[4][1]["refs"][-1] == 3391
    assert awk_data[6][2]["refs"][-1] == 676


def test_awkward_array_tvector2_array_forth():

    awk_data = None

    with uproot.open(skhep_testdata.data_path("uproot-HZZ-objects.root"))[
        "events/MET"
    ] as tree:
        interp = uproot.interpretation.identify.interpretation_of(tree, {}, False)
        interp._forth = True
        awk_data = tree.array(interp, library="ak")
    assert awk_data[0]["fX"] == pytest.approx(5.912771224975586)
    assert awk_data[4]["fY"] == pytest.approx(-1.3100523948669434)
    assert awk_data[1200]["fX"] == pytest.approx(1.9457910060882568)


def test_awkward_vector_tstring():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["vector_tstring"].array(library="ak")) == [
            ["one"],
            ["one", "two"],
            ["one", "two", "three"],
            ["one", "two", "three", "four"],
            ["one", "two", "three", "four", "five"],
        ]


def test_awkward_vector_vector_int32():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["vector_vector_int32"].array(library="ak")) == [
            [[1]],
            [[1], [1, 2]],
            [[1], [1, 2], [1, 2, 3]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4], [1, 2, 3, 4, 5]],
        ]


def test_awkward_vector_vector_string():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["vector_vector_string"].array(library="ak")) == [
            [["one"]],
            [["one"], ["one", "two"]],
            [["one"], ["one", "two"], ["one", "two", "three"]],
            [
                ["one"],
                ["one", "two"],
                ["one", "two", "three"],
                ["one", "two", "three", "four"],
            ],
            [
                ["one"],
                ["one", "two"],
                ["one", "two", "three"],
                ["one", "two", "three", "four"],
                ["one", "two", "three", "four", "five"],
            ],
        ]


def test_awkward_vector_vector_tstring():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["vector_vector_tstring"].array(library="ak")) == [
            [["one"]],
            [["one"], ["one", "two"]],
            [["one"], ["one", "two"], ["one", "two", "three"]],
            [
                ["one"],
                ["one", "two"],
                ["one", "two", "three"],
                ["one", "two", "three", "four"],
            ],
            [
                ["one"],
                ["one", "two"],
                ["one", "two", "three"],
                ["one", "two", "three", "four"],
                ["one", "two", "three", "four", "five"],
            ],
        ]


def test_awkward_set_int32():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["set_int32"].array(library="ak")) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]


def test_awkward_vector_set_int32():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["vector_set_int32"].array(library="ak")) == [
            [[1]],
            [[1], [1, 2]],
            [[1], [1, 2], [1, 2, 3]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4], [1, 2, 3, 4, 5]],
        ]


def test_awkward_vector_set_string():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["vector_set_string"].array(library="ak")) == [
            [["one"]],
            [["one"], ["one", "two"]],
            [["one"], ["one", "two"], ["one", "three", "two"]],
            [
                ["one"],
                ["one", "two"],
                ["one", "three", "two"],
                ["four", "one", "three", "two"],
            ],
            [
                ["one"],
                ["one", "two"],
                ["one", "three", "two"],
                ["four", "one", "three", "two"],
                ["five", "four", "one", "three", "two"],
            ],
        ]


def test_awkward_set_string():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["set_string"].array(library="ak")) == [
            ["one"],
            ["one", "two"],
            ["one", "three", "two"],
            ["four", "one", "three", "two"],
            ["five", "four", "one", "three", "two"],
        ]


def test_awkward_map_int32_int16():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_int32_int16"].array(library="ak"))
        assert awkward.to_list(keys) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]
        assert awkward.to_list(values) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]


def test_awkward_map_int32_vector_int16():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_int32_vector_int16"].array(library="ak"))
        assert awkward.to_list(keys) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]
        assert awkward.to_list(values) == [
            [[1]],
            [[1], [1, 2]],
            [[1], [1, 2], [1, 2, 3]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4], [1, 2, 3, 4, 5]],
        ]


def test_awkward_map_int32_vector_string():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(
            tree["map_int32_vector_string"].array(library="ak")
        )
        assert awkward.to_list(keys) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]
        assert awkward.to_list(values) == [
            [["one"]],
            [["one"], ["one", "two"]],
            [["one"], ["one", "two"], ["one", "two", "three"]],
            [
                ["one"],
                ["one", "two"],
                ["one", "two", "three"],
                ["one", "two", "three", "four"],
            ],
            [
                ["one"],
                ["one", "two"],
                ["one", "two", "three"],
                ["one", "two", "three", "four"],
                ["one", "two", "three", "four", "five"],
            ],
        ]


def test_awkward_map_int32_set_int16():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_int32_set_int16"].array(library="ak"))
        assert awkward.to_list(keys) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]
        assert awkward.to_list(values) == [
            [[1]],
            [[1], [1, 2]],
            [[1], [1, 2], [1, 2, 3]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4], [1, 2, 3, 4, 5]],
        ]


def test_awkward_map_int32_set_string():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_int32_set_string"].array(library="ak"))
        assert awkward.to_list(keys) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]
        assert awkward.to_list(values) == [
            [["one"]],
            [["one"], ["one", "two"]],
            [["one"], ["one", "two"], ["one", "three", "two"]],
            [
                ["one"],
                ["one", "two"],
                ["one", "three", "two"],
                ["four", "one", "three", "two"],
            ],
            [
                ["one"],
                ["one", "two"],
                ["one", "three", "two"],
                ["four", "one", "three", "two"],
                ["five", "four", "one", "three", "two"],
            ],
        ]


def test_awkward_map_string_int16():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_string_int16"].array(library="ak"))
        assert awkward.to_list(keys) == [
            ["one"],
            ["one", "two"],
            ["one", "three", "two"],
            ["four", "one", "three", "two"],
            ["five", "four", "one", "three", "two"],
        ]
        assert awkward.to_list(values) == [
            [1],
            [1, 2],
            [1, 3, 2],
            [4, 1, 3, 2],
            [5, 4, 1, 3, 2],
        ]


def test_awkward_map_string_vector_int16():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(
            tree["map_string_vector_int16"].array(library="ak")
        )
        assert awkward.to_list(keys) == [
            ["one"],
            ["one", "two"],
            ["one", "three", "two"],
            ["four", "one", "three", "two"],
            ["five", "four", "one", "three", "two"],
        ]
        assert awkward.to_list(values) == [
            [[1]],
            [[1], [1, 2]],
            [[1], [1, 2, 3], [1, 2]],
            [[1, 2, 3, 4], [1], [1, 2, 3], [1, 2]],
            [[1, 2, 3, 4, 5], [1, 2, 3, 4], [1], [1, 2, 3], [1, 2]],
        ]


def test_awkward_map_string_vector_string():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(
            tree["map_string_vector_string"].array(library="ak")
        )
        assert awkward.to_list(keys) == [
            ["one"],
            ["one", "two"],
            ["one", "three", "two"],
            ["four", "one", "three", "two"],
            ["five", "four", "one", "three", "two"],
        ]
        assert awkward.to_list(values) == [
            [["one"]],
            [["one"], ["one", "two"]],
            [["one"], ["one", "two", "three"], ["one", "two"]],
            [
                ["one", "two", "three", "four"],
                ["one"],
                ["one", "two", "three"],
                ["one", "two"],
            ],
            [
                ["one", "two", "three", "four", "five"],
                ["one", "two", "three", "four"],
                ["one"],
                ["one", "two", "three"],
                ["one", "two"],
            ],
        ]


def test_awkward_map_string_set_int16():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_string_set_int16"].array(library="ak"))
        assert awkward.to_list(keys) == [
            ["one"],
            ["one", "two"],
            ["one", "three", "two"],
            ["four", "one", "three", "two"],
            ["five", "four", "one", "three", "two"],
        ]
        assert awkward.to_list(values) == [
            [[1]],
            [[1], [1, 2]],
            [[1], [1, 2, 3], [1, 2]],
            [[1, 2, 3, 4], [1], [1, 2, 3], [1, 2]],
            [[1, 2, 3, 4, 5], [1, 2, 3, 4], [1], [1, 2, 3], [1, 2]],
        ]


def test_awkward_map_string_set_string():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_string_set_string"].array(library="ak"))
        assert awkward.to_list(keys) == [
            ["one"],
            ["one", "two"],
            ["one", "three", "two"],
            ["four", "one", "three", "two"],
            ["five", "four", "one", "three", "two"],
        ]
        assert awkward.to_list(values) == [
            [["one"]],
            [["one"], ["one", "two"]],
            [["one"], ["one", "three", "two"], ["one", "two"]],
            [
                ["four", "one", "three", "two"],
                ["one"],
                ["one", "three", "two"],
                ["one", "two"],
            ],
            [
                ["five", "four", "one", "three", "two"],
                ["four", "one", "three", "two"],
                ["one"],
                ["one", "three", "two"],
                ["one", "two"],
            ],
        ]


def test_awkward_map_int32_vector_vector_int16():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(
            tree["map_int32_vector_vector_int16"].array(library="ak")
        )
        assert awkward.to_list(keys) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]
        assert awkward.to_list(values) == [
            [[[1]]],
            [[[1]], [[1], [1, 2]]],
            [[[1]], [[1], [1, 2]], [[1], [1, 2], [1, 2, 3]]],
            [
                [[1]],
                [[1], [1, 2]],
                [[1], [1, 2], [1, 2, 3]],
                [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4]],
            ],
            [
                [[1]],
                [[1], [1, 2]],
                [[1], [1, 2], [1, 2, 3]],
                [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4]],
                [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4], [1, 2, 3, 4, 5]],
            ],
        ]


def test_awkward_map_string_string():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_string_string"].array(library="ak"))
        assert awkward.to_list(keys) == [
            ["one"],
            ["one", "two"],
            ["one", "three", "two"],
            ["four", "one", "three", "two"],
            ["five", "four", "one", "three", "two"],
        ]
        assert awkward.to_list(values) == [
            ["ONE"],
            ["ONE", "TWO"],
            ["ONE", "THREE", "TWO"],
            ["FOUR", "ONE", "THREE", "TWO"],
            ["FIVE", "FOUR", "ONE", "THREE", "TWO"],
        ]


def test_awkward_map_string_tstring():

    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_string_tstring"].array(library="ak"))
        assert awkward.to_list(keys) == [
            ["one"],
            ["one", "two"],
            ["one", "three", "two"],
            ["four", "one", "three", "two"],
            ["five", "four", "one", "three", "two"],
        ]
        assert awkward.to_list(values) == [
            ["ONE"],
            ["ONE", "TWO"],
            ["ONE", "THREE", "TWO"],
            ["FOUR", "ONE", "THREE", "TWO"],
            ["FIVE", "FOUR", "ONE", "THREE", "TWO"],
        ]



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines an :doc:`uproot.interpretation.Interpretation` for
several numerical types:

* :doc:`uproot.interpretation.numerical.AsDtype`: numbers, which can simply be
  described as a ``numpy.dtype``.
* :doc:`uproot.interpretation.numerical.AsDtypeInPlace`: a predefined array
  into which data may be overwritten.
* :doc:`uproot.interpretation.numerical.AsDouble32`: ROOT's ``Double32_t``
  packed data type.
* :doc:`uproot.interpretation.numerical.AsFloat16`: ROOT's ``Float16_t``
  packed data type.
* :doc:`uproot.interpretation.numerical.AsSTLBits`: an ``std::bitset<N>``
  for some ``N``.
"""


import numpy

import uproot


def _dtype_shape(dtype):
    shape = ()
    while dtype.subdtype is not None:
        dtype, s = dtype.subdtype
        shape = shape + s
    return dtype, shape


class Numerical(uproot.interpretation.Interpretation):
    """
    Abstract superclass of numerical interpretations, including

    * :doc:`uproot.interpretation.numerical.AsDtype`
    * :doc:`uproot.interpretation.numerical.AsSTLBits`
    * :doc:`uproot.interpretation.numerical.TruncatedNumerical`
    """

    def _wrap_almost_finalized(self, array):
        return array

    def final_array(
        self,
        basket_arrays,
        entry_start,
        entry_stop,
        entry_offsets,
        library,
        branch,
        options,
    ):
        self.hook_before_final_array(
            basket_arrays=basket_arrays,
            entry_start=entry_start,
            entry_stop=entry_stop,
            entry_offsets=entry_offsets,
            library=library,
            branch=branch,
        )

        if entry_start >= entry_stop:
            output = self._prepare_output(library, length=0)

        else:
            length = 0
            start = entry_offsets[0]
            for _, stop in enumerate(entry_offsets[1:]):
                if start <= entry_start and entry_stop <= stop:
                    length += entry_stop - entry_start
                elif start <= entry_start < stop:
                    length += stop - entry_start
                elif start <= entry_stop <= stop:
                    length += entry_stop - start
                elif entry_start < stop and start <= entry_stop:
                    length += stop - start
                start = stop

            output = self._prepare_output(library, length)

            start = entry_offsets[0]
            for basket_num, stop in enumerate(entry_offsets[1:]):
                if start <= entry_start and entry_stop <= stop:
                    local_start = entry_start - start
                    local_stop = entry_stop - start
                    basket_array = basket_arrays[basket_num]
                    output[:] = basket_array[local_start:local_stop]

                elif start <= entry_start < stop:
                    local_start = entry_start - start
                    local_stop = stop - start
                    basket_array = basket_arrays[basket_num]
                    output[: stop - entry_start] = basket_array[local_start:local_stop]

                elif start <= entry_stop <= stop:
                    local_start = 0
                    local_stop = entry_stop - start
                    basket_array = basket_arrays[basket_num]
                    output[start - entry_start :] = basket_array[local_start:local_stop]

                elif entry_start < stop and start <= entry_stop:
                    basket_array = basket_arrays[basket_num]
                    output[start - entry_start : stop - entry_start] = basket_array

                start = stop

        self.hook_before_library_finalize(
            basket_arrays=basket_arrays,
            entry_start=entry_start,
            entry_stop=entry_stop,
            entry_offsets=entry_offsets,
            library=library,
            branch=branch,
            output=output,
        )

        output = self._wrap_almost_finalized(output)

        output = library.finalize(
            output, branch, self, entry_start, entry_stop, options
        )

        self.hook_after_final_array(
            basket_arrays=basket_arrays,
            entry_start=entry_start,
            entry_stop=entry_stop,
            entry_offsets=entry_offsets,
            library=library,
            branch=branch,
            output=output,
        )

        return output

    def _prepare_output(self, library, length):
        """
        Prepare the output array in which the data is stored.

        In this default implementation, just create an empty array from the library but specializations might re-use an existing array (ex: :doc:`uproot.interpretation.numerical.AsDtypeInPlace`:)
        """
        output = library.empty((length,), self.to_dtype)
        return output


_numpy_byteorder_to_cache_key = {
    "!": "B",
    ">": "B",
    "<": "L",
    "|": "L",
    "=": "B" if numpy.dtype(">f8").isnative else "L",
}

_dtype_kind_itemsize_to_typename = {
    ("b", 1): "bool",
    ("i", 1): "int8_t",
    ("u", 1): "uint8_t",
    ("i", 2): "int16_t",
    ("u", 2): "uint16_t",
    ("i", 4): "int32_t",
    ("u", 4): "uint32_t",
    ("i", 8): "int64_t",
    ("u", 8): "uint64_t",
    ("f", 4): "float",
    ("f", 8): "double",
}


class AsDtype(Numerical):
    """
    Args:
        from_dtype (``numpy.dtype`` or its constructor argument): Data type to
            *assume* of the raw but uncompressed bytes in the ``TBasket``.
            Usually big-endian; may include named fields and a shape.
        to_dtype (None, ``numpy.dtype``, or its constructor argument): Data
            type to *convert* the data into. Usually native-endian; may include
            named fields and a shape. If None, ``to_dtype`` will be set to the
            native-endian equivalent of ``from_dtype``.

    Interpretation for any array that can be fully described as a
    ``numpy.dtype``.
    """

    def __init__(self, from_dtype, to_dtype=None):
        self._from_dtype = numpy.dtype(from_dtype)
        if to_dtype is None:
            self._to_dtype = self._from_dtype.newbyteorder("=")
        else:
            self._to_dtype = numpy.dtype(to_dtype)

    def __repr__(self):
        if self._to_dtype == self._from_dtype.newbyteorder("="):
            return f"AsDtype({str(self._from_dtype)!r})"
        else:
            return f"AsDtype({str(self._from_dtype)!r}, {str(self._to_dtype)!r})"

    def __eq__(self, other):
        return (
            type(other) is AsDtype
            and self._from_dtype == other._from_dtype
            and self._to_dtype == other._to_dtype
        )

    @property
    def from_dtype(self):
        """
        Data type to expect of the raw but uncompressed bytes in the
        ``TBasket`` data. Usually big-endian; may include named fields and a
        shape.

        Named fields (``dtype.names``) can be used to construct a NumPy
        `structured array <https://numpy.org/doc/stable/user/basics.rec.html>`__.

        A shape (``dtype.shape``) can be used to construct a fixed-size array
        for each entry. (Not applicable to variable-length lists! See
        :doc:`uproot.interpretation.jagged.AsJagged`.) The finalized array's
        ``array.shape[1:] == dtype.shape``.
        """
        return self._from_dtype

    @property
    def to_dtype(self):
        """
        Data type to convert the data into. Usually the native-endian
        equivalent of :ref:`uproot.interpretation.numerical.AsDtype.from_dtype`;
        may include named fields and a shape.

        Named fields (``dtype.names``) can be used to construct a NumPy
        `structured array <https://numpy.org/doc/stable/user/basics.rec.html>`__.

        A shape (``dtype.shape``) can be used to construct a fixed-size array
        for each entry. (Not applicable to variable-length lists! See
        :doc:`uproot.interpretation.jagged.AsJagged`.) The finalized array's
        ``array.shape[1:] == dtype.shape``.
        """
        return self._to_dtype

    @property
    def itemsize(self):
        """
        Number of bytes per item of
        :ref:`uproot.interpretation.numerical.AsDtype.from_dtype`.

        This number of bytes includes the fields and shape, like
        ``dtype.itemsize`` in NumPy.
        """
        return self._from_dtype.itemsize

    @property
    def inner_shape(self):
        _, s = _dtype_shape(self._from_dtype)
        return s

    @property
    def numpy_dtype(self):
        return self._to_dtype

    def awkward_form(
        self,
        file,
        context=None,
        index_format="i64",
        header=False,
        tobject_header=False,
        breadcrumbs=(),
    ):
        context = self._make_context(
            context, index_format, header, tobject_header, breadcrumbs
        )
        awkward = uproot.extras.awkward()
        d, s = _dtype_shape(self._to_dtype)
        out = uproot._util.awkward_form(d, file, context)
        for size in s[::-1]:
            out = awkward.forms.RegularForm(out, size)
        return out

    @property
    def cache_key(self):
        def form(dtype, name):
            d, s = _dtype_shape(dtype)
            return "{}{}{}({}{})".format(
                _numpy_byteorder_to_cache_key[d.byteorder],
                d.kind,
                d.itemsize,
                ",".join(repr(x) for x in s),
                name,
            )

        if self.from_dtype.names is None:
            from_dtype = form(self.from_dtype, "")
        else:
            from_dtype = (
                "["
                + ",".join(
                    form(self.from_dtype[n], "," + repr(n))
                    for n in self.from_dtype.names
                )
                + "]"
            )

        if self.to_dtype.names is None:
            to_dtype = form(self.to_dtype, "")
        else:
            to_dtype = (
                "["
                + ",".join(
                    form(self.to_dtype[n], "," + repr(n)) for n in self.to_dtype.names
                )
                + "]"
            )

        return f"{type(self).__name__}({from_dtype},{to_dtype})"

    @property
    def typename(self):
        def form(dtype):
            d, s = _dtype_shape(dtype)
            return _dtype_kind_itemsize_to_typename[d.kind, d.itemsize] + "".join(
                "[" + str(dim) + "]" for dim in s
            )

        if self.from_dtype.names is None:
            return form(self.from_dtype)
        else:
            return (
                "struct {"
                + " ".join(
                    f"{form(self.from_dtype[n])} {n};" for n in self.from_dtype.names
                )
                + "}"
            )

    def basket_array(
        self,
        data,
        byte_offsets,
        basket,
        branch,
        context,
        cursor_offset,
        library,
        options,
    ):
        self.hook_before_basket_array(
            data=data,
            byte_offsets=byte_offsets,
            basket=basket,
            branch=branch,
            context=context,
            cursor_offset=cursor_offset,
            library=library,
            options=options,
        )

        dtype, shape = _dtype_shape(self._from_dtype)
        try:
            output = data.view(dtype).reshape((-1,) + shape)
        except ValueError as err:
            raise ValueError(
                """basket {} in tree/branch {} has the wrong number of bytes ({}) """
                """for interpretation {}
in file {}""".format(
                    basket.basket_num,
                    branch.object_path,
                    len(data),
                    self,
                    branch.file.file_path,
                )
            ) from err

        self.hook_after_basket_array(
            data=data,
            byte_offsets=byte_offsets,
            basket=basket,
            branch=branch,
            context=context,
            output=output,
            cursor_offset=cursor_offset,
            library=library,
            options=options,
        )

        return output

    def reshape(self, shape):
        d, s = _dtype_shape(self._from_dtype)
        self._from_dtype = numpy.dtype((d, shape))
        d, s = _dtype_shape(self._to_dtype)
        self._to_dtype = numpy.dtype((d, shape))

    def inplace(self, array):
        """
        Returns a AsDtypeInPlace version of self in order to fill the given array in place.

        Example usage :
        ```
        var = np.zeros(N, dtype=np.float32)
        b = uproot.openn('afile.root')['treename']['varname']
        b.array(library='np', interpretation=b.interpretation.inplace(var) )
        ```
        """
        return AsDtypeInPlace(array, self._from_dtype)


class AsDtypeInPlace(AsDtype):
    """
    Like :doc:`uproot.interpretation.numerical.AsDtype`, but a given array is
    filled in-place, rather than creating a new output array.
    """

    def __init__(self, array, from_dtype):
        self._to_fill = array
        self._from_dtype = from_dtype
        self._to_dtype = numpy.dtype(array.dtype)

    def _prepare_output(self, library, length):
        """
        Specialized version of _prepare_output : re-use our target array kept in self._to_fill.
        """
        if library.name != "np":
            raise TypeError(
                "AsDtypeInPlace can only be used with library 'np', not '{}'".format(
                    library.name
                )
            )

        output = self._to_fill.view(self.to_dtype)

        if length > len(output):
            raise ValueError(
                "Requesting to fill an array of size {} (type {}) with input of size {} (type {})".format(
                    len(output), self._to_dtype, length, self._from_dtype
                )
            )
        return output[:length]


class AsSTLBits(Numerical):
    """
    Interpretation for ``std::bitset``.
    """

    def __init__(self):
        raise NotImplementedError

    @property
    def itemsize(self):
        return self._num_bytes + 4


class TruncatedNumerical(Numerical):
    """
    Abstract superclass for interpretations that truncate the range and
    granularity of the real number line to pack data into fewer bits.

    Subclasses are

    * :doc:`uproot.interpretation.numerical.AsDouble32`
    * :doc:`uproot.interpretation.numerical.AsFloat16`
    """

    @property
    def low(self):
        """
        Lower bound on the range of real numbers this type can express.
        """
        return self._low

    @property
    def high(self):
        """
        Upper bound on the range of real numbers this type can express.
        """
        return self._high

    @property
    def num_bits(self):
        """
        Number of bytes into which to pack these data.
        """
        return self._num_bits

    @property
    def from_dtype(self):
        """
        The ``numpy.dtype`` of the raw but uncompressed data.

        May be a
        `structured array <https://numpy.org/doc/stable/user/basics.rec.html>`__
        of ``"exponent"`` and ``"mantissa"`` or an integer.
        """
        if self.is_truncated:
            return numpy.dtype(({"exponent": (">u1", 0), "mantissa": (">u2", 1)}, ()))
        else:
            return numpy.dtype(">u4")

    @property
    def itemsize(self):
        """
        Number of bytes in
        :ref:`uproot.interpretation.numerical.TruncatedNumerical.from_dtype`.
        """
        return self.from_dtype.itemsize

    @property
    def to_dims(self):
        """
        The ``dtype.shape`` of the ``to_dtype``.
        """
        return self._to_dims

    @property
    def is_truncated(self):
        """
        If True (:ref:`uproot.interpretation.numerical.TruncatedNumerical.low`
        and :ref:`uproot.interpretation.numerical.TruncatedNumerical.high` are
        both ``0``), the data are truly truncated.
        """
        return self._low == self._high == 0.0

    def __repr__(self):
        args = [repr(self._low), repr(self._high), repr(self._num_bits)]
        if self._to_dims != ():
            args.append(f"to_dims={self._to_dims!r}")
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and self._low == other._low
            and self._high == other._high
            and self._num_bits == other._num_bits
            and self._to_dims == other._to_dims
        )

    @property
    def numpy_dtype(self):
        return self.to_dtype

    @property
    def cache_key(self):
        return "{}({},{},{},{})".format(
            type(self).__name__, self._low, self._high, self._num_bits, self._to_dims
        )

    def basket_array(
        self,
        data,
        byte_offsets,
        basket,
        branch,
        context,
        cursor_offset,
        library,
        options,
    ):
        self.hook_before_basket_array(
            data=data,
            byte_offsets=byte_offsets,
            basket=basket,
            branch=branch,
            context=context,
            cursor_offset=cursor_offset,
            library=library,
            options=options,
        )

        try:
            raw = data.view(self.from_dtype)
        except ValueError as err:
            raise ValueError(
                """basket {} in tree/branch {} has the wrong number of bytes ({}) """
                """for interpretation {} (expecting raw array of {})
in file {}""".format(
                    basket.basket_num,
                    branch.object_path,
                    len(data),
                    self,
                    repr(self._from_dtype),
                    branch.file.file_path,
                )
            ) from err

        if self.is_truncated:
            exponent = raw["exponent"].astype(numpy.int32)
            mantissa = raw["mantissa"].astype(numpy.int32)

            exponent <<= 23
            exponent |= (mantissa & ((1 << (self.num_bits + 1)) - 1)) << (
                23 - self.num_bits
            )
            sign = ((1 << (self.num_bits + 1)) & mantissa != 0) * -2 + 1

            output = exponent.view(numpy.float32) * sign

            d, s = _dtype_shape(self.to_dtype)
            output = output.astype(d).reshape((-1,) + s)

        else:
            d, s = _dtype_shape(self.to_dtype)
            output = raw.astype(d).reshape((-1,) + s)
            numpy.multiply(
                output,
                float(self._high - self._low) / (1 << self._num_bits),
                out=output,
            )
            numpy.add(output, self.low, out=output)

        self.hook_after_basket_array(
            data=data,
            byte_offsets=byte_offsets,
            basket=basket,
            branch=branch,
            context=context,
            cursor_offset=cursor_offset,
            library=library,
            raw=raw,
            output=output,
            options=options,
        )

        return output


class AsDouble32(TruncatedNumerical):
    """
    Args:
        low (float): Lower bound on the range of expressible values.
        high (float): Upper bound on the range of expressible values.
        num_bits (int): Number of bits in the representation.
        to_dims (tuple of ints): Shape of
            :ref:`uproot.interpretation.numerical.AsDouble32.to_dtype`.

    Interpretation for ROOT's ``Double32_t`` type.
    """

    def __init__(self, low, high, num_bits, to_dims=()):
        self._low = low
        self._high = high
        self._num_bits = num_bits
        self._to_dims = to_dims

        if not uproot._util.isint(num_bits) or not 2 <= num_bits <= 32:
            raise TypeError("num_bits must be an integer between 2 and 32 (inclusive)")
        if high <= low and not self.is_truncated:
            raise ValueError(f"high ({high}) must be strictly greater than low ({low})")

    @property
    def to_dtype(self):
        """
        The ``numpy.dtype`` of the output array.

        A shape (``dtype.shape``) can be used to construct a fixed-size array
        for each entry. (Not applicable to variable-length lists! See
        :doc:`uproot.interpretation.jagged.AsJagged`.) The finalized array's
        ``array.shape[1:] == dtype.shape``.
        """
        return numpy.dtype((numpy.float64, self.to_dims))

    @property
    def typename(self):
        return "Double32_t" + "".join("[" + str(dim) + "]" for dim in self._to_dims)

    def awkward_form(
        self,
        file,
        context=None,
        index_format="i64",
        header=False,
        tobject_header=False,
        breadcrumbs=(),
    ):
        context = self._make_context(
            context, index_format, header, tobject_header, breadcrumbs
        )
        awkward = uproot.extras.awkward()
        out = awkward.forms.NumpyForm("float64")
        for size in self._to_dims[::-1]:
            out = awkward.forms.RegularForm(out, size)
        return out


class AsFloat16(TruncatedNumerical):
    """
    Args:
        low (float): Lower bound on the range of expressible values.
        high (float): Upper bound on the range of expressible values.
        num_bits (int): Number of bits in the representation.
        to_dims (tuple of ints): Shape of
            :ref:`uproot.interpretation.numerical.AsFloat16.to_dtype`.

    Interpretation for ROOT's ``Float16_t`` type.
    """

    def __init__(self, low, high, num_bits, to_dims=()):
        self._low = low
        self._high = high
        self._num_bits = num_bits
        self._to_dims = to_dims

        if not uproot._util.isint(num_bits) or not 2 <= num_bits <= 32:
            raise TypeError("num_bits must be an integer between 2 and 32 (inclusive)")
        if high <= low and not self.is_truncated:
            raise ValueError(f"high ({high}) must be strictly greater than low ({low})")

    @property
    def to_dtype(self):
        """
        The ``numpy.dtype`` of the output array.

        A shape (``dtype.shape``) can be used to construct a fixed-size array
        for each entry. (Not applicable to variable-length lists! See
        :doc:`uproot.interpretation.jagged.AsJagged`.) The finalized array's
        ``array.shape[1:] == dtype.shape``.
        """
        return numpy.dtype((numpy.float32, self.to_dims))

    @property
    def typename(self):
        return "Float16_t" + "".join("[" + str(dim) + "]" for dim in self._to_dims)

    def awkward_form(
        self,
        file,
        context=None,
        index_format="i64",
        header=False,
        tobject_header=False,
        breadcrumbs=(),
    ):
        context = self._make_context(
            context, index_format, header, tobject_header, breadcrumbs
        )
        awkward = uproot.extras.awkward()
        out = awkward.forms.NumpyForm("float32")
        for size in self._to_dims[::-1]:
            out = awkward.forms.RegularForm(out, size)
        return out



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import json
import sys

import numpy
import pytest
import skhep_testdata

import uproot

from_ROOT = {}

from_ROOT["TH1F"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "TH1F",
  "fTitle" : "",
  "fCheckSum" : 3642409091,
  "fClassVersion" : 2,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : [{
      "_typename" : "TStreamerBase",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "TH1",
      "fTitle" : "1-Dim histogram base class",
      "fType" : 0,
      "fSize" : 0,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 1063172259, 0, 0, 0],
      "fTypeName" : "BASE",
      "fBaseVersion" : 7
    }, {
      "_typename" : "TStreamerBase",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "TArrayF",
      "fTitle" : "Array of floats",
      "fType" : 0,
      "fSize" : 0,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 1510733553, 0, 0, 0],
      "fTypeName" : "BASE",
      "fBaseVersion" : 1
    }]
  }
}"""
)

from_ROOT["TH1"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "TH1",
  "fTitle" : "",
  "fCheckSum" : 1063172259,
  "fClassVersion" : 7,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : [{
      "_typename" : "TStreamerBase",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "TNamed",
      "fTitle" : "The basis for a named object (name, title)",
      "fType" : 67,
      "fSize" : 0,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, -541636036, 0, 0, 0],
      "fTypeName" : "BASE",
      "fBaseVersion" : 1
    }, {
      "_typename" : "TStreamerBase",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "TAttLine",
      "fTitle" : "Line attributes",
      "fType" : 0,
      "fSize" : 0,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, -1811462839, 0, 0, 0],
      "fTypeName" : "BASE",
      "fBaseVersion" : 2
    }, {
      "_typename" : "TStreamerBase",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "TAttFill",
      "fTitle" : "Fill area attributes",
      "fType" : 0,
      "fSize" : 0,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, -2545006, 0, 0, 0],
      "fTypeName" : "BASE",
      "fBaseVersion" : 2
    }, {
      "_typename" : "TStreamerBase",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "TAttMarker",
      "fTitle" : "Marker attributes",
      "fType" : 0,
      "fSize" : 0,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 689802220, 0, 0, 0],
      "fTypeName" : "BASE",
      "fBaseVersion" : 2
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fNcells",
      "fTitle" : "number of bins(1D), cells (2D) +U\/Overflows",
      "fType" : 3,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "int"
    }, {
      "_typename" : "TStreamerObject",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fXaxis",
      "fTitle" : "X axis descriptor",
      "fType" : 61,
      "fSize" : 216,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TAxis"
    }, {
      "_typename" : "TStreamerObject",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fYaxis",
      "fTitle" : "Y axis descriptor",
      "fType" : 61,
      "fSize" : 216,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TAxis"
    }, {
      "_typename" : "TStreamerObject",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fZaxis",
      "fTitle" : "Z axis descriptor",
      "fType" : 61,
      "fSize" : 216,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TAxis"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fBarOffset",
      "fTitle" : "(1000*offset) for bar charts or legos",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fBarWidth",
      "fTitle" : "(1000*width) for bar charts or legos",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fEntries",
      "fTitle" : "Number of entries",
      "fType" : 8,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "double"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fTsumw",
      "fTitle" : "Total Sum of weights",
      "fType" : 8,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "double"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fTsumw2",
      "fTitle" : "Total Sum of squares of weights",
      "fType" : 8,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "double"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fTsumwx",
      "fTitle" : "Total Sum of weight*X",
      "fType" : 8,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "double"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fTsumwx2",
      "fTitle" : "Total Sum of weight*X*X",
      "fType" : 8,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "double"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fMaximum",
      "fTitle" : "Maximum value for plotting",
      "fType" : 8,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "double"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fMinimum",
      "fTitle" : "Minimum value for plotting",
      "fType" : 8,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "double"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fNormFactor",
      "fTitle" : "Normalization factor",
      "fType" : 8,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "double"
    }, {
      "_typename" : "TStreamerObjectAny",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fContour",
      "fTitle" : "Array to display contour levels",
      "fType" : 62,
      "fSize" : 24,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TArrayD"
    }, {
      "_typename" : "TStreamerObjectAny",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fSumw2",
      "fTitle" : "Array of sum of squares of weights",
      "fType" : 62,
      "fSize" : 24,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TArrayD"
    }, {
      "_typename" : "TStreamerString",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fOption",
      "fTitle" : "histogram options",
      "fType" : 65,
      "fSize" : 24,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TString"
    }, {
      "_typename" : "TStreamerObjectPointer",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fFunctions",
      "fTitle" : "->Pointer to list of functions (fits and user)",
      "fType" : 63,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TList*"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fBufferSize",
      "fTitle" : "fBuffer size",
      "fType" : 6,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "int"
    }, {
      "_typename" : "TStreamerBasicPointer",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fBuffer",
      "fTitle" : "[fBufferSize] entry buffer",
      "fType" : 48,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "double*",
      "fCountVersion" : 7,
      "fCountName" : "fBufferSize",
      "fCountClass" : "TH1"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fBinStatErrOpt",
      "fTitle" : "option for bin statistical errors",
      "fType" : 3,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TH1::EBinErrorOpt"
    }]
  }
}"""
)

from_ROOT["TNamed"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "TNamed",
  "fTitle" : "",
  "fCheckSum" : 3753331260,
  "fClassVersion" : 1,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : [{
      "_typename" : "TStreamerBase",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "TObject",
      "fTitle" : "Basic ROOT object",
      "fType" : 66,
      "fSize" : 0,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, -1877229523, 0, 0, 0],
      "fTypeName" : "BASE",
      "fBaseVersion" : 1
    }, {
      "_typename" : "TStreamerString",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fName",
      "fTitle" : "object identifier",
      "fType" : 65,
      "fSize" : 24,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TString"
    }, {
      "_typename" : "TStreamerString",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fTitle",
      "fTitle" : "object title",
      "fType" : 65,
      "fSize" : 24,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TString"
    }]
  }
}"""
)

from_ROOT["TObject"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "TObject",
  "fTitle" : "",
  "fCheckSum" : 2417737773,
  "fClassVersion" : 1,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : [{
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fUniqueID",
      "fTitle" : "object unique identifier",
      "fType" : 13,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "unsigned int"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fBits",
      "fTitle" : "bit field status word",
      "fType" : 15,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "unsigned int"
    }]
  }
}"""
)

from_ROOT["TAttLine"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "TAttLine",
  "fTitle" : "",
  "fCheckSum" : 2483504457,
  "fClassVersion" : 2,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : [{
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fLineColor",
      "fTitle" : "Line color",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fLineStyle",
      "fTitle" : "Line style",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fLineWidth",
      "fTitle" : "Line width",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }]
  }
}"""
)

from_ROOT["TAttFill"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "TAttFill",
  "fTitle" : "",
  "fCheckSum" : 4292422290,
  "fClassVersion" : 2,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : [{
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fFillColor",
      "fTitle" : "Fill area color",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fFillStyle",
      "fTitle" : "Fill area style",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }]
  }
}"""
)

from_ROOT["TAttMarker"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "TAttMarker",
  "fTitle" : "",
  "fCheckSum" : 689802220,
  "fClassVersion" : 2,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : [{
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fMarkerColor",
      "fTitle" : "Marker color",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fMarkerStyle",
      "fTitle" : "Marker style",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fMarkerSize",
      "fTitle" : "Marker size",
      "fType" : 5,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "float"
    }]
  }
}"""
)

from_ROOT["TAxis"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "TAxis",
  "fTitle" : "",
  "fCheckSum" : 1514761840,
  "fClassVersion" : 10,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : [{
      "_typename" : "TStreamerBase",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "TNamed",
      "fTitle" : "The basis for a named object (name, title)",
      "fType" : 67,
      "fSize" : 0,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, -541636036, 0, 0, 0],
      "fTypeName" : "BASE",
      "fBaseVersion" : 1
    }, {
      "_typename" : "TStreamerBase",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "TAttAxis",
      "fTitle" : "Axis attributes",
      "fType" : 0,
      "fSize" : 0,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 1550843710, 0, 0, 0],
      "fTypeName" : "BASE",
      "fBaseVersion" : 4
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fNbins",
      "fTitle" : "Number of bins",
      "fType" : 3,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "int"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fXmin",
      "fTitle" : "low edge of first bin",
      "fType" : 8,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "double"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fXmax",
      "fTitle" : "upper edge of last bin",
      "fType" : 8,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "double"
    }, {
      "_typename" : "TStreamerObjectAny",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fXbins",
      "fTitle" : "Bin edges array in X",
      "fType" : 62,
      "fSize" : 24,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TArrayD"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fFirst",
      "fTitle" : "first bin to display",
      "fType" : 3,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "int"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fLast",
      "fTitle" : "last bin to display",
      "fType" : 3,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "int"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fBits2",
      "fTitle" : "second bit status word",
      "fType" : 12,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "unsigned short"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fTimeDisplay",
      "fTitle" : "on\/off displaying time values instead of numerics",
      "fType" : 18,
      "fSize" : 1,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "bool"
    }, {
      "_typename" : "TStreamerString",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fTimeFormat",
      "fTitle" : "Date&time format, ex: 09\/12\/99 12:34:00",
      "fType" : 65,
      "fSize" : 24,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TString"
    }, {
      "_typename" : "TStreamerObjectPointer",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fLabels",
      "fTitle" : "List of labels",
      "fType" : 64,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "THashList*"
    }, {
      "_typename" : "TStreamerObjectPointer",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fModLabs",
      "fTitle" : "List of modified labels",
      "fType" : 64,
      "fSize" : 8,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TList*"
    }]
  }
}"""
)

from_ROOT["TAttAxis"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "TAttAxis",
  "fTitle" : "",
  "fCheckSum" : 1550843710,
  "fClassVersion" : 4,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : [{
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fNdivisions",
      "fTitle" : "Number of divisions(10000*n3 + 100*n2 + n1)",
      "fType" : 3,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "int"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fAxisColor",
      "fTitle" : "Color of the line axis",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fLabelColor",
      "fTitle" : "Color of labels",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fLabelFont",
      "fTitle" : "Font for labels",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fLabelOffset",
      "fTitle" : "Offset of labels",
      "fType" : 5,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "float"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fLabelSize",
      "fTitle" : "Size of labels",
      "fType" : 5,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "float"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fTickLength",
      "fTitle" : "Length of tick marks",
      "fType" : 5,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "float"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fTitleOffset",
      "fTitle" : "Offset of axis title",
      "fType" : 5,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "float"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fTitleSize",
      "fTitle" : "Size of axis title",
      "fType" : 5,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "float"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fTitleColor",
      "fTitle" : "Color of axis title",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fTitleFont",
      "fTitle" : "Font for axis title",
      "fType" : 2,
      "fSize" : 2,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "short"
    }]
  }
}"""
)

from_ROOT["THashList"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "THashList",
  "fTitle" : "",
  "fCheckSum" : 3430828481,
  "fClassVersion" : 0,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : [{
      "_typename" : "TStreamerBase",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "TList",
      "fTitle" : "Doubly linked list",
      "fType" : 0,
      "fSize" : 0,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 1774568379, 0, 0, 0],
      "fTypeName" : "BASE",
      "fBaseVersion" : 5
    }]
  }
}"""
)

from_ROOT["TList"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "TList",
  "fTitle" : "",
  "fCheckSum" : 1774568379,
  "fClassVersion" : 5,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : [{
      "_typename" : "TStreamerBase",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "TSeqCollection",
      "fTitle" : "Sequenceable collection ABC",
      "fType" : 0,
      "fSize" : 0,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, -60015674, 0, 0, 0],
      "fTypeName" : "BASE",
      "fBaseVersion" : 0
    }]
  }
}"""
)

from_ROOT["TSeqCollection"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "TSeqCollection",
  "fTitle" : "",
  "fCheckSum" : 4234951622,
  "fClassVersion" : 0,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : [{
      "_typename" : "TStreamerBase",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "TCollection",
      "fTitle" : "Collection abstract base class",
      "fType" : 0,
      "fSize" : 0,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 1474546588, 0, 0, 0],
      "fTypeName" : "BASE",
      "fBaseVersion" : 3
    }]
  }
}"""
)

from_ROOT["TCollection"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "TCollection",
  "fTitle" : "",
  "fCheckSum" : 1474546588,
  "fClassVersion" : 3,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : [{
      "_typename" : "TStreamerBase",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "TObject",
      "fTitle" : "Basic ROOT object",
      "fType" : 66,
      "fSize" : 0,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, -1877229523, 0, 0, 0],
      "fTypeName" : "BASE",
      "fBaseVersion" : 1
    }, {
      "_typename" : "TStreamerString",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fName",
      "fTitle" : "name of the collection",
      "fType" : 65,
      "fSize" : 24,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "TString"
    }, {
      "_typename" : "TStreamerBasicType",
      "fUniqueID" : 0,
      "fBits" : 0,
      "fName" : "fSize",
      "fTitle" : "number of elements in collection",
      "fType" : 3,
      "fSize" : 4,
      "fArrayLength" : 0,
      "fArrayDim" : 0,
      "fMaxIndex" : [0, 0, 0, 0, 0],
      "fTypeName" : "int"
    }]
  }
}"""
)

from_ROOT["TString"] = json.loads(
    r"""{
  "_typename" : "TStreamerInfo",
  "fUniqueID" : 0,
  "fBits" : 0,
  "fName" : "TString",
  "fTitle" : "",
  "fCheckSum" : 95257,
  "fClassVersion" : 2,
  "fElements" : {
    "_typename" : "TObjArray",
    "name" : "TObjArray",
    "arr" : []
  }
}"""
)


def drop_fbits(x):
    if isinstance(x, dict):
        return {k: drop_fbits(v) for k, v in x.items() if k != "fBits"}
    elif isinstance(x, list):
        return list(drop_fbits(v) for v in x)
    else:
        return x


def test():
    with uproot.open(skhep_testdata.data_path("uproot-histograms.root")) as f:
        streamers = f.file.streamers
        assert len(streamers) == 14

        assert drop_fbits(from_ROOT["TH1F"]) == drop_fbits(
            streamers["TH1F"][2].tojson()
        )
        assert drop_fbits(from_ROOT["TH1"]) == drop_fbits(streamers["TH1"][7].tojson())
        assert drop_fbits(from_ROOT["TNamed"]) == drop_fbits(
            streamers["TNamed"][1].tojson()
        )
        assert drop_fbits(from_ROOT["TObject"]) == drop_fbits(
            streamers["TObject"][1].tojson()
        )
        assert drop_fbits(from_ROOT["TAttLine"]) == drop_fbits(
            streamers["TAttLine"][2].tojson()
        )
        assert drop_fbits(from_ROOT["TAttFill"]) == drop_fbits(
            streamers["TAttFill"][2].tojson()
        )
        assert drop_fbits(from_ROOT["TAttMarker"]) == drop_fbits(
            streamers["TAttMarker"][2].tojson()
        )
        assert drop_fbits(from_ROOT["TAxis"]) == drop_fbits(
            streamers["TAxis"][10].tojson()
        )
        assert drop_fbits(from_ROOT["TAttAxis"]) == drop_fbits(
            streamers["TAttAxis"][4].tojson()
        )
        assert drop_fbits(from_ROOT["THashList"]) == drop_fbits(
            streamers["THashList"][0].tojson()
        )
        assert drop_fbits(from_ROOT["TList"]) == drop_fbits(
            streamers["TList"][5].tojson()
        )
        assert drop_fbits(from_ROOT["TSeqCollection"]) == drop_fbits(
            streamers["TSeqCollection"][0].tojson()
        )
        assert drop_fbits(from_ROOT["TCollection"]) == drop_fbits(
            streamers["TCollection"][3].tojson()
        )
        assert drop_fbits(from_ROOT["TString"]) == drop_fbits(
            streamers["TString"][2].tojson()
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy as np
import pytest
import skhep_testdata

import uproot

pytest.importorskip("pandas")
pytest.importorskip("awkward_pandas")


def test():
    for arrays in uproot.iterate(
        [skhep_testdata.data_path("uproot-HZZ-uncompressed.root") + ":events"] * 2,
        ["Muon_Px", "Jet_Px", "MET_px"],
        library="pd",
    ):
        pass



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import pytest
import skhep_testdata

import uproot
from uproot.models.TTable import Model_TTable


class Model_StIOEvent(uproot.model.Model):
    def read_members(self, chunk, cursor, context, file):
        self._bases.append(
            uproot.models.TObject.Model_TObject.read(
                chunk,
                cursor,
                context,
                file,
                self._file,
                self._parent,
                concrete=self.concrete,
            )
        )

        self._members["fObj"] = uproot.deserialization.read_object_any(
            chunk, cursor, context, file, self._file, self.concrete
        )


@pytest.fixture(scope="module")
def datafile():
    yield skhep_testdata.data_path("uproot-issue-418.root")


@pytest.fixture
def custom_classes():
    yield dict(
        uproot.classes,
        StIOEvent=Model_StIOEvent,
        St_g2t_event=Model_TTable,
        St_particle=Model_TTable,
        St_g2t_vertex=Model_TTable,
        St_g2t_track=Model_TTable,
        St_g2t_pythia=Model_TTable,
        St_g2t_tpc_hit=Model_TTable,
        St_g2t_ctf_hit=Model_TTable,
        St_g2t_emc_hit=Model_TTable,
        St_g2t_vpd_hit=Model_TTable,
        St_dst_bfc_status=Model_TTable,
    )


@pytest.fixture
def geant_branch(datafile, custom_classes):
    with uproot.open(datafile, custom_classes=custom_classes) as f:
        yield f["geantBranch.0000000123.0000000322"]


def test_geant_dot_root(geant_branch):
    items = {obj.all_members["fName"]: obj for obj in geant_branch.members["fObj"]}
    assert items["g2t_pythia"].data["subprocess_id"] == 1



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import io
import os

import numpy as np
import pytest
import skhep_testdata

import uproot

ROOT = pytest.importorskip("ROOT")


def test(tmp_path):
    filename = os.path.join(tmp_path, "whatever.root")

    with uproot.recreate(filename) as f:
        f["t"] = {"b": np.array([1, 2, 3], np.int64)}

    with uproot.open(filename) as f:
        output = io.StringIO()
        f.file.show_streamers(stream=output)
        assert len(output.getvalue()) > 100
        assert len(f.file.streamers) == 24
        assert f["t/b"].array(library="np").tolist() == [1, 2, 3]

    f = ROOT.TFile(filename)
    t = f.Get("t")
    assert [x.b for x in t] == [1, 2, 3]
